package gs.test.v0.api

import scala.quoted.*

/** Represents a specific location in source code.
  *
  * ## Credit
  *
  * This code was taken directly from https://github.com/tpolecat/SourcePos
  * (also under the MIT License).
  *
  * ## Usage
  *
  * {{{
  *   def example()(using sp: SourcePosition): Unit =
  *     println(s"Called from $sp")
  * }}}
  *
  * @param file
  *   The specific source file being referenced.
  * @param line
  *   The line number within the source file.
  */
case class SourcePosition(
  file: String,
  line: Int
):
  override def toString(): String = s"$file:$line"

object SourcePosition:

  /** Implicit, inlined function to receive the source position at the inlined
    * location.
    */
  implicit inline def instance: SourcePosition =
    ${ sourcePositionImpl }

  /** Implementation for getting the [[SourcePosition]] using macros.
    *
    * @param context
    *   The current quotation context.
    * @return
    *   Expression that produces the source position at the point of macro
    *   expansion.
    */
  def sourcePositionImpl(
    using
    context: Quotes
  ): Expr[SourcePosition] =
    val rootPosition = context.reflect.Position.ofMacroExpansion
    val file         = Expr(rootPosition.sourceFile.path)
    val line         = Expr(rootPosition.startLine + 1)
    '{ SourcePosition($file, $line) }

end SourcePosition
