package gs.test.v0.api

import cats.~>
import cats.Applicative
import cats.data.EitherT
import cats.effect.kernel.Async
import cats.effect.kernel.Ref
import cats.effect.kernel.Resource
import cats.effect.syntax.all.*
import cats.syntax.all.*
import java.net.URI
import natchez.*

sealed trait UnitOfWork[F[_]]:

  def work(
    span: Span[F]
  ): F[Either[TestFailure, Any]]

object UnitOfWork:

  type Traced[F[_]] = natchez.Trace[F] ?=> F[Either[TestFailure, Any]]

  type TracedT[F[_]] = natchez.Trace[F] ?=> EitherT[F, TestFailure, Any]

  /** Instantiate a new [[UnitOfWork]] with the given function that requires a
    * `Span[F]` as input.
    *
    * @param uow
    *   The unit of work implementation.
    * @return
    *   The new [[UnitOfWork]] instance.
    */
  def apply[F[_]: Async](
    unitOfWork: Traced[F]
  ): UnitOfWork[F] =
    new UnitOfWork[F] {

      override def work(span: Span[F]): F[Either[TestFailure, Any]] =
        makeInternalTrace[F](span).flatMap { trace =>
          given Trace[F] = trace
          unitOfWork
        }

    }

  def applyT[F[_]: Async](
    unitOfWork: TracedT[F]
  ): UnitOfWork[F] =
    new UnitOfWork[F] {

      override def work(span: Span[F]): F[Either[TestFailure, Any]] =
        makeInternalTrace[F](span).flatMap { trace =>
          given Trace[F] = trace
          unitOfWork.value
        }

    }

  private def makeInternalTrace[F[_]: Async](sourceSpan: Span[F])
    : F[natchez.Trace[F]] =
    Ref.of[F, Span[F]](sourceSpan).map(src => new InternalTrace[F](src))

  // Copied from the Natchez ioTrace
  private class InternalTrace[F[_]: Async](
    val src: Ref[F, Span[F]]
  ) extends natchez.Trace[F]:

    def put(fields: (String, TraceValue)*): F[Unit] =
      src.get.flatMap(_.put(fields*))

    def log(fields: (String, TraceValue)*): F[Unit] =
      src.get.flatMap(_.log(fields*))

    def log(event: String): F[Unit] =
      src.get.flatMap(_.log(event))

    def attachError(
      err: Throwable,
      fields: (String, TraceValue)*
    ): F[Unit] =
      src.get.flatMap(_.attachError(err, fields*))

    def kernel: F[Kernel] = src.get.flatMap(_.kernel)

    def spanR(
      name: String,
      options: Span.Options = Span.Options.Defaults
    ): Resource[F, F ~> F] =
      for {
        parent <- Resource.eval(src.get)
        child  <- parent.span(name, options)
      } yield new (F ~> F) {
        def apply[A](fa: F[A]): F[A] =
          src.get.flatMap { old =>
            src
              .set(child)
              .bracket(_ => fa.onError { case e => child.attachError(e) })(_ =>
                src.set(old)
              )
          }

      }

    def span[A](
      name: String,
      options: Span.Options = Span.Options.Defaults
    )(
      k: F[A]
    ): F[A] =
      spanR(name, options).use(_(k))

    def traceId: F[Option[String]] =
      src.get.flatMap(_.traceId)

    override def spanId(
      implicit
      F: Applicative[F]
    ): F[Option[String]] =
      src.get.flatMap(_.spanId)

    override def traceUri: F[Option[URI]] = src.get.flatMap(_.traceUri)

  end InternalTrace

end UnitOfWork
