package gs.test.v0.api

import cats.data.Kleisli
import cats.effect.Async
import cats.effect.IO
import gs.test.v0.api.{Tag => GsTag}
import munit.*
import natchez.Span

class GroupImplementationTests extends FunSuite:
  import GroupImplementationTests.*

  test("should support a group with a simple, pure, test") {
    val g1    = new G1[[A] =>> Kleisli[IO, Span[IO], A]]
    val group = g1.compile()
    assertEquals(group.name, TestGroupDefinition.Name("G1"))
    assertEquals(group.documentation, None)
    assertEquals(group.testTags, List.empty)
    assertEquals(group.testMarkers, List.empty)
    assertEquals(group.beforeGroup, None)
    assertEquals(group.afterGroup, None)
    assertEquals(group.beforeEachTest, None)
    assertEquals(group.afterEachTest, None)
    assertEquals(group.tests.size, 1)

    group.tests match
      case t :: Nil =>
        assertEquals(t.name, TestDefinition.Name("simple"))
        assertEquals(t.permanentId, Ids.T1)
        assertEquals(t.documentation, None)
        assertEquals(t.tags, List.empty)
        assertEquals(t.markers, List.empty)
        assertEquals(t.iterations, TestIterations.One)
      case _ => fail("Unexpected number of defined tests.")
  }

  test("should support a group with all values set") {
    val g2    = new G2[[A] =>> Kleisli[IO, Span[IO], A]]
    val group = g2.compile()
    assertEquals(group.name, TestGroupDefinition.Name("G2"))
    assertEquals(group.documentation, Some("docs"))
    assertEquals(group.testTags, List(GsTag("tag")))
    assertEquals(group.testMarkers, List(Marker.Ignored))
    assertEquals(group.beforeGroup.isDefined, true)
    assertEquals(group.afterGroup.isDefined, true)
    assertEquals(group.beforeEachTest.isDefined, true)
    assertEquals(group.afterEachTest.isDefined, true)
    assertEquals(group.tests.size, 1)

    group.tests match
      case t :: Nil =>
        assertEquals(t.name, TestDefinition.Name("inherit from group"))
        assertEquals(t.permanentId, Ids.T2)
        assertEquals(t.documentation, None)
        assertEquals(t.tags, List(GsTag("tag")))
        assertEquals(t.markers, List(Marker.Ignored))
        assertEquals(t.iterations, TestIterations.One)
      case _ => fail("Unexpected number of defined tests.")
  }

  test("should support a simple group with a configured test") {
    val g3    = new G3[[A] =>> Kleisli[IO, Span[IO], A]]
    val group = g3.compile()
    assertEquals(group.name, TestGroupDefinition.Name("G3"))
    assertEquals(group.documentation, None)
    assertEquals(group.testTags, List.empty)
    assertEquals(group.testMarkers, List.empty)
    assertEquals(group.beforeGroup, None)
    assertEquals(group.afterGroup, None)
    assertEquals(group.beforeEachTest, None)
    assertEquals(group.afterEachTest, None)
    assertEquals(group.tests.size, 1)

    group.tests match
      case t :: Nil =>
        assertEquals(t.name, TestDefinition.Name("configure test"))
        assertEquals(t.permanentId, Ids.T3)
        assertEquals(t.documentation, Some("docs"))
        assertEquals(t.tags, List(tag"tag1", tag"tag2"))
        assertEquals(t.markers, List(Marker.Ignored))
        assertEquals(t.iterations, TestIterations(2))
      case _ => fail("Unexpected number of defined tests.")
  }

object GroupImplementationTests:

  object Ids:

    val T1: PermanentId = pid"t1"
    val T2: PermanentId = pid"t2"
    val T3: PermanentId = pid"t3"
    val T4: PermanentId = pid"t4"

  end Ids

  class G1[F[_]: Async] extends TestGroup[F]:
    override def name: String = "G1"
    test(Ids.T1, "simple").pure(Right(()))
  end G1

  class G2[F[_]: Async] extends TestGroup[F]:

    override def name: String =
      "G2"

    override def documentation: Option[String] =
      Some("docs")

    override def tags: List[GsTag] =
      List(GsTag("tag"))

    override def markers: List[Marker] =
      List(Marker.Ignored)

    beforeGroup(Async[F].unit)
    afterGroup(Async[F].unit)
    beforeEachTest(Async[F].unit)
    afterEachTest(Async[F].unit)

    test(Ids.T2, "inherit from group").pure(Right(()))
  end G2

  class G3[F[_]: Async] extends TestGroup[F]:
    override def name: String = "G3"

    test(Ids.T3, "configure test")
      .document("docs")
      .tagged(tag"tag1", tag"tag2")
      .marked(Marker.Ignored)
      .iterate(TestIterations(2))
      .pure(Right(()))

  end G3

  class G4[F[_]: Async] extends TestGroup[F]:
    override def name: String = "G4"

    // TODO: Make test entrypoint and test Trace[F]
    test(Ids.T4, "Effectful test").effectful {
      ???
    }

end GroupImplementationTests
